package ru.sicamp.sicamphelper.generator

import com.squareup.kotlinpoet.asTypeName
import mu.KLogging
import nl.devhaan.kotlinpoetdsl.classes.clazz
import nl.devhaan.kotlinpoetdsl.constructorBuilder.buildConstructor
import nl.devhaan.kotlinpoetdsl.constructorBuilder.constructor
import nl.devhaan.kotlinpoetdsl.constructorBuilder.thiz
import nl.devhaan.kotlinpoetdsl.files.file
import nl.devhaan.kotlinpoetdsl.of
import nl.devhaan.kotlinpoetdsl.packaged
import nl.devhaan.kotlinpoetdsl.public
import nl.devhaan.kotlinpoetdsl.valOf
import ru.sicamp.sicamphelper.db.entity.Room
import ru.sicamp.sicamphelper.db.entity.User
import java.io.File
import kotlin.reflect.KClass
import kotlin.reflect.full.memberProperties

class InfoGenerator {
    fun <T : Any> generateInfo(kClass: KClass<T>) {
        val className = "${kClass.simpleName}Info"
        logger.info("Beginning generating $className")
        val file = File("C:\\Users\\Andrey\\IdeaProjects\\sicamp-helper\\build\\generated\\source")
        file(GENERATED_PACKAGE, className) {
            val constructorParameters = kClass.infoVariables()
            val secondaryConstructor = buildConstructor {
                val entityName = "entity"
                constructor(entityName of kClass).thiz(
                    *constructorParameters.map {
                        "$entityName.${it.name}"
                    }.toTypedArray()
                ) {}
            }
            logger.info("Found ${constructorParameters.size} parameters")
            clazz(className, *constructorParameters.toTypedArray()) {
                public.constructor(secondaryConstructor)
            }.packaged(GENERATED_PACKAGE)
        }.writeTo(file)
    }

    private fun KClass<*>.infoPropertiesWithAnnotations() = memberProperties.mapNotNull { property ->
        property.annotations.firstOrNull { it is Info }?.let { annotation ->
            property to annotation
        }
    }.also { logger.info("Found ${it.size} properties") }

    private fun KClass<*>.infoVariables() = infoPropertiesWithAnnotations()
        .map { (property, annotation) ->
            logger.info("Getting type for ${property.name}")
            val typeName = property.returnType.asTypeName()
            property.name.valOf(
                if (annotation is Info) {
                    if (property.returnType.isMarkedNullable) {
                        typeName.asNullable()
                    } else {
                        typeName
                    }
                } else {
                    error("Annotation $annotation is not of type Info")
                }
            )
        }

    companion object : KLogging() {
        const val GENERATED_PACKAGE = "ru.sicamp.helper.generated"
    }
}

@Target(AnnotationTarget.CLASS)
annotation class GeneratorTarget

@Target(AnnotationTarget.PROPERTY)
@Retention(AnnotationRetention.SOURCE)
annotation class Info

@Target(AnnotationTarget.PROPERTY)
@Retention(AnnotationRetention.SOURCE)
annotation class Dto

@GeneratorTarget
class A(
    @Info
    val a: String? = "a",
    @Info
    val c: String = "c",
    @Dto
    val b: String = "b"
)


fun main() {
    InfoGenerator().generateInfo(User::class)
    InfoGenerator().generateInfo(Room::class)
}